#include <catch2/catch.hpp>

#include <language/ast/ASTBuilder.hpp>
#include <language/ast/ASTNodeDataTypeBuilder.hpp>
#include <language/ast/ASTNodeDeclarationToAffectationConverter.hpp>
#include <language/ast/ASTNodeExpressionBuilder.hpp>
#include <language/ast/ASTNodeTypeCleaner.hpp>
#include <language/ast/ASTSymbolTableBuilder.hpp>
#include <language/node_processor/OStreamProcessor.hpp>
#include <utils/Demangle.hpp>

#include <pegtl/string_input.hpp>

#include <sstream>

void
_replaceOStream(ASTNode& node, std::ostringstream& sout)
{
  if (node.is_type<language::cout_kw>() or node.is_type<language::cerr_kw>()) {
    node.m_node_processor = std::make_unique<OStreamProcessor>(node, sout);
  } else {
    for (auto& child_node : node.children) {
      _replaceOStream(*child_node, sout);
    }
  }
}

#define CHECK_OSTREAM_EXPRESSION_RESULT(data, expected_value)  \
  {                                                            \
    TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"}; \
    auto ast = ASTBuilder::build(input);                       \
                                                               \
    ASTSymbolTableBuilder{*ast};                               \
    ASTNodeDataTypeBuilder{*ast};                              \
                                                               \
    ASTNodeDeclarationToAffectationConverter{*ast};            \
    ASTNodeTypeCleaner<language::var_declaration>{*ast};       \
                                                               \
    ASTNodeExpressionBuilder{*ast};                            \
    ExecutionPolicy exec_policy;                               \
                                                               \
    std::ostringstream sout;                                   \
    _replaceOStream(*ast, sout);                               \
                                                               \
    ast->execute(exec_policy);                                 \
                                                               \
    REQUIRE(sout.str() == expected_value);                     \
  }

#define CHECK_OSTREAM_EXPRESSION_THROWS(data, expected_error)                                   \
  {                                                                                             \
    static_assert(std::is_same_v<std::decay_t<decltype(data)>, std::string_view>);              \
    static_assert((std::is_same_v<std::decay_t<decltype(expected_error)>, std::string_view>) or \
                  (std::is_same_v<std::decay_t<decltype(expected_error)>, std::string>));       \
                                                                                                \
    TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};                                  \
    auto ast = ASTBuilder::build(input);                                                        \
                                                                                                \
    ASTSymbolTableBuilder{*ast};                                                                \
    ASTNodeDataTypeBuilder{*ast};                                                               \
                                                                                                \
    ASTNodeDeclarationToAffectationConverter{*ast};                                             \
    ASTNodeTypeCleaner<language::var_declaration>{*ast};                                        \
    ASTNodeTypeCleaner<language::fct_declaration>{*ast};                                        \
                                                                                                \
    REQUIRE_THROWS_WITH(ASTNodeExpressionBuilder{*ast}, expected_error);                        \
  }

// clazy:excludeall=non-pod-global-static

TEST_CASE("OStreamProcessor", "[language]")
{
  SECTION("cout")
  {
    CHECK_OSTREAM_EXPRESSION_RESULT(R"(cout << 2;)", "2");
    CHECK_OSTREAM_EXPRESSION_RESULT(R"(cout << true;)", "true");
    CHECK_OSTREAM_EXPRESSION_RESULT(R"(cout << false;)", "false");
    CHECK_OSTREAM_EXPRESSION_RESULT(R"(cout << "x=" << 2 << "\n";)", "x=2\n");
  }

  SECTION("cerr")
  {
    CHECK_OSTREAM_EXPRESSION_RESULT(R"(cerr << 2;)", "2");
    CHECK_OSTREAM_EXPRESSION_RESULT(R"(cerr << true;)", "true");
    CHECK_OSTREAM_EXPRESSION_RESULT(R"(cerr << false;)", "false");
    CHECK_OSTREAM_EXPRESSION_RESULT(R"(cerr << "x=" << 2 << "\n";)", "x=2\n");
  }

  SECTION("runtime error")
  {
    std::string_view data_type = R"(
let f : R -> R, x -> 2;
cerr << "f=" << f << "\n";
)";

    std::string error_msg = "invalid argument, cannot print a 'function'";
    CHECK_OSTREAM_EXPRESSION_THROWS(data_type, error_msg);
  }
}
